
# line 2 "parse.y"
/* pathalias -- by steve bellovin, as told to peter honeyman */
#ifndef lint
static char	*sccsid = "@(#)parse.y	9.1 87/10/04";
#endif /* lint */

#include "def.h"

/* exports */
extern void yyerror();

/* imports */
extern p_node addnode();
extern p_node addprivate();
extern void fixprivate(), alias(), deadlink();
extern p_link addlink();
#ifndef TMPFILES
#define getnode(x) x
#define getlink(y) y
#else  /*TMPFILES*/
extern node *getnode();
extern link *getlink();
#endif /*TMPFILES*/
extern int optind;
extern char *Cfile, *Netchars, **Argv;
extern int Argc;
extern long Lineno;

/* privates */
STATIC void fixnet();
STATIC int yylex(), yywrap(), getword();
static int Scanstate = NEWLINE;	/* scanner (yylex) state */

/* flags for ys_flags */
#define TERMINAL 1

# line 38 "parse.y"
typedef union  {
	p_node	y_node;
	Cost	y_cost;
	char	y_net;
	char	*y_name;
	struct {
		p_node ys_node;
		Cost ys_cost;
		short ys_flag;
		char ys_net;
		char ys_dir;
	} y_s;
} YYSTYPE;
# define SITE 257
# define HOST 258
# define COST 259
# define NET 260
# define EOL 261
# define PRIVATE 262
# define DEAD 263
#define yyclearin yychar = -1
#define yyerrok yyerrflag = 0
extern int yychar;
extern short yyerrflag;
#ifndef YYMAXDEPTH
#define YYMAXDEPTH 150
#endif
YYSTYPE yylval, yyval;
# define YYERRCODE 256

# line 189 "parse.y"


void
yyerror(s)
	char *s;
{
	/* a concession to bsd error(1) */
	if (Cfile)
		fprintf(stderr, "\"%s\", ", Cfile);
	else
		fprintf(stderr, "%s: ", Argv[0]);
	fprintf(stderr, "line %ld: %s\n", Lineno, s);
}

/*
 * patch in the costs of getting on/off the network.
 *
 * for each network member on netlist, add links:
 *	network -> member	cost = 0;
 *	member -> network	cost = parameter.
 *
 * if network and member both require gateways, assume network
 * is a gateway to member (but not v.v., to avoid such travesties
 * as topaz!seismo.css.gov.edu.rutgers).
 *
 * note that members can have varying costs to a network, by suitable
 * multiple declarations.  this is a feechur, albeit a useless one.
 */
STATIC void
fixnet(network, nlist, cost, netchar, netdir)
	register p_node network;
	p_node nlist;
	Cost cost;
	char netchar, netdir;
{	register p_node member;
	register p_node nextnet;
	p_link l;

	getnode(network)->n_flag |= NNET;

	/* now insert the links */
	for (member = nlist ; member; member = nextnet) {
		/* network -> member, cost is 0 */
		l = addlink(network, member, (Cost) 0, netchar, netdir);
		if (GATEWAYED(getnode(network)) && GATEWAYED(getnode(member)))
			getlink(l)->l_flag |= LGATEWAY;

		/* member -> network, cost is parameter */
		(void) addlink(member, network, cost, netchar, netdir);
		nextnet = getnode(member)->n_net;
		getnode(member)->n_net = 0;	/* clear for later use */
	}
}

/* scanner */

#define QUOTE '"'
#define STR_EQ(s1, s2) (s1[2] == s2[2] && strcmp(s1, s2) == 0)
#define NLRETURN() {Scanstate = NEWLINE; return EOL;}

static struct ctable {
	char *cname;
	Cost cval;
} ctable[] = {
	/* ordered by frequency of appearance in a "typical" dataset */
	{"DIRECT", 200},
	{"DEMAND", 300},
	{"DAILY", 5000},
	{"HOURLY", 500},
	{"DEDICATED", 95},
	{"EVENING", 1800},
	{"LOCAL", 25},
	{"LOW", 5},	/* baud rate, quality penalty */
	{"DEAD", INF/2},
	{"POLLED", 5000},
	{"WEEKLY", 30000},
	{"HIGH", -5},	/* baud rate, quality bonus */
	{"FAST", -80},	/* high speed (>= 9.6 kbps) modem */
	/* deprecated */
	{"ARPA", 100},
	{"DIALED", 300},
	{0, 0}
};

STATIC int
yylex()
{	static char retbuf[128];	/* for return to yacc part */
	register int c;
	register char *buf = retbuf;
	register struct ctable *ct;
	register Cost cost;
	char errbuf[128];

	if (feof(stdin) && yywrap())
		return EOF;

	/* count lines, skip over space and comments */
	if ((c = getchar()) == EOF)
		NLRETURN();
    
continuation:
	while (c == ' ' || c == '\t')
		if ((c = getchar()) == EOF)
			NLRETURN();

	if (c == '#')
		while ((c = getchar()) != '\n')
			if (c == EOF)
				NLRETURN();

	/* scan token */
	if (c == '\n') {
		Lineno++;
		if ((c = getchar()) != EOF) {
			if (c == ' ' || c == '\t')
				goto continuation;
			ungetc(c, stdin);
		}
		NLRETURN();
	}

	switch(Scanstate) {
	case COSTING:
		if (isdigit(c)) {
			cost = c - '0';
			for (c = getchar(); isdigit(c); c = getchar())
				cost = (cost * 10) + c - '0';
			ungetc(c, stdin);
			yylval.y_cost = cost;
			return COST;
		}

		if (getword(buf, c) == 0) {
			for (ct = ctable; ct->cname; ct++)
				if (STR_EQ(buf, ct->cname)) {
					yylval.y_cost = ct->cval;
					return COST;
				}
			sprintf(errbuf, "unknown cost (%s), using default", buf);
			yyerror(errbuf);
			yylval.y_cost = DEFCOST;
			return COST;
		}

		return c;	/* pass the buck */

	case NEWLINE:
		Scanstate = OTHER;
		if (getword(buf, c) != 0)
			return c;
		/* `private' (but not `"private"')? */
		if (c == 'p' && STR_EQ(buf, "private"))
			return PRIVATE;
		/* `dead' (but not `"dead"')? */
		if (c == 'd' && STR_EQ(buf, "dead"))
			return DEAD;

		yylval.y_name = buf;
		return HOST;
	}

	if (getword(buf, c) == 0) {
		yylval.y_name = buf;
		return SITE;
	}

	if (index(Netchars, c)) {
		yylval.y_net = c;
		return NET;
	}

	return c;
}

/*
 * fill str with the next word in [0-9A-Za-z][-._0-9A-Za-z]+ or a quoted
 * string that contains no newline.  return -1 on failure or EOF, 0 o.w.
 */ 
STATIC int
getword(str, c)
	register char *str;
	register int c;
{
	if (c == QUOTE) {
		while ((c = getchar()) != QUOTE) {
			if (c == '\n') {
				yyerror("newline in quoted string\n");
				ungetc(c, stdin);
				return -1;
			}
			if (c == EOF) {
				yyerror("EOF in quoted string\n");
				return -1;
			}
			*str++ = c;
		}
		*str = 0;
		return 0;
	}

	/* host name must start with alphanumeric or `.' */
	if (!isalnum(c) && c != '.')
		return -1;

yymore:
	do {
		*str++ = c;
		c = getchar();
	} while (isalnum(c) || c == '.' || c == '_');

	if (c == '-' && Scanstate != COSTING)
		goto yymore;

	ungetc(c, stdin);
	*str = 0;
	return 0;
}

STATIC int
yywrap()
{	char errbuf[100];

	fixprivate();	/* munge private host definitions */
	Lineno = 1;
	while (optind < Argc) {
		if (freopen((Cfile = Argv[optind++]), "r", stdin) != 0)
			return 0;
		sprintf(errbuf, "%s: %s", Argv[0], Cfile);
		perror(errbuf);
	}
	freopen("/dev/null", "r", stdin);
	return -1;
}
short yyexca[] ={
-1, 1,
	0, -1,
	-2, 0,
-1, 45,
	260, 24,
	-2, 34,
	};
# define YYNPROD 52
# define YYLAST 235
short yyact[]={

  12,  26,  22,   3,  10,  11,  34,  26,  17,  15,
  32,  20,  19,  18,  13,  60,  57,  36,  40,  26,
  67,  45,  42,  50,  38,  30,   2,  62,  62,  56,
  54,  51,  28,  27,  58,  52,  31,  44,  46,  49,
  32,  71,  41,  83,  71,  69,  72,  70,  78,  72,
  71,  69,  23,  70,  21,  72,  68,  48,   8,   7,
   1,  43,   9,   6,  39,   5,  47,   4,   0,   0,
  29,   0,   0,   0,   0,   0,   0,  37,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,  63,   0,   0,  65,  73,  66,  64,  74,   0,
   0,  76,   0,   0,  79,  80,  81,  82,  77,  61,
  55,  53,  84,   0,  85,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
  33,   0,   0,  35,   0,   0,   0,   0,   0,   0,
  42,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
   0,   0,   0,   0,   0,   0,   0,   0,  25,   0,
   0,  24,   0,   0,  25,   0,   0,  24,   0,   0,
   0,   0,   0,   0,   0,   0,  25,   0,   0,   0,
   0,   0,   0,   0,   0,  16,  14,   0,   0,   0,
  75,   0,   0,   0,  59 };
short yypact[]={

-230,-258,-247,-1000, -35, -36,-248,-249,-250, -59,
 -90, -91,-1000,-1000,-1000, -53,-1000,-232,-1000,-1000,
-1000,   0,-117,-243, -41,-1000,-233,-107,-236,   0,
-1000,-1000,-1000,-1000,-234, -92,-1000,-1000, -27, -14,
-1000,-1000,-1000, -15,-1000,-1000,-244,-1000, -25, -16,
-1000,-234,-1000,-1000,-235,-1000,-236,-237,   8,-1000,
 -25, -30,-237, -17,-1000,-1000,-1000,-1000,   7, -25,
 -25, -25, -25,   2,-1000,   0,-1000,   0,-1000,  -1,
  -1,-1000,-1000,-1000,-1000,-1000 };
short yypgo[]={

   0,  54,  52,  67,  65,  64,  63,  39,  62,  42,
  38,  37,  61,  36,  34,  60,  59,  58,  57,  56 };
short yyr1[]={

   0,  15,  15,  15,  15,  15,  15,  15,  15,   3,
   3,   3,   4,   4,   4,   6,   6,   6,  16,  16,
  17,   8,   8,   8,  10,   2,   2,   1,   1,   1,
   9,   5,   5,   5,  11,  11,  12,  12,  12,   7,
   7,   7,  13,  18,  19,  13,  14,  14,  14,  14,
  14,  14 };
short yyr2[]={

   0,   0,   2,   3,   3,   3,   3,   3,   2,   3,
   4,   2,   3,   3,   2,   6,   7,   7,   4,   3,
   4,   1,   1,   1,   1,   1,   3,   1,   2,   2,
   1,   1,   3,   2,   1,   3,   1,   3,   2,   1,
   3,   2,   0,   0,   0,   5,   1,   3,   3,   3,
   3,   3 };
short yychk[]={

-1000, -15, 256, 261,  -3,  -4,  -6, -16, -17,  -8,
 262, 263, 258, 261, 261,  44, 261,  44, 261, 261,
 261,  -1,  61,  -2, 260, 257,  60, 123, 123,  -1,
 257, -13,  40, 257, 123, 260, 260,  -2, 257,  -5,
 125,  -9, 257, -12, -11, 257, -10, -13, -18,  -7,
 257, 123,  62, 125,  44, 125,  44, 260, -14, 259,
  40, 125,  44,  -7,  -9, -11, -10, 257, -19,  43,
  45,  42,  47, -14, -13, 260, -10, 125,  41, -14,
 -14, -14, -14,  41, -13, -13 };
short yydef[]={

   1,  -2,   0,   2,   0,   0,   0,   0,   0,   0,
  22,  23,  21,   8,   3,  11,   4,  14,   5,   6,
   7,  42,   0,  27,   0,  25,   0,   0,   0,  42,
  13,   9,  43,  12,   0,   0,  29,  28,   0,   0,
  19,  31,  30,   0,  36,  -2,   0,  10,   0,   0,
  39,   0,  26,  18,  33,  20,  38,   0,  44,  46,
   0,  42,  41,   0,  32,  37,  35,  24,   0,   0,
   0,   0,   0,   0,  15,  42,  40,  42,  45,  48,
  49,  50,  51,  47,  17,  16 };
#if	!defined(lint) && defined(DOSCCS)
static char yaccpar_sccsid[] = "@(#)yaccpar	4.1	(Berkeley)	2/11/83";
#endif

#
# define YYFLAG -1000
# define YYERROR goto yyerrlab
# define YYACCEPT return(0)
# define YYABORT return(1)

/*	parser for yacc output	*/

#ifdef YYDEBUG
int yydebug = 0; /* 1 for debugging */
#endif
YYSTYPE yyv[YYMAXDEPTH]; /* where the values are stored */
int yychar = -1; /* current input token number */
int yynerrs = 0;  /* number of errors */
short yyerrflag = 0;  /* error recovery flag */

yyparse() {

	short yys[YYMAXDEPTH];
	short yyj, yym;
	register YYSTYPE *yypvt;
	register short yystate, *yyps, yyn;
	register YYSTYPE *yypv;
	register short *yyxi;

	yystate = 0;
	yychar = -1;
	yynerrs = 0;
	yyerrflag = 0;
	yyps= &yys[-1];
	yypv= &yyv[-1];

 yystack:    /* put a state and value onto the stack */

#ifdef YYDEBUG
	if( yydebug  ) printf( "state %d, char 0%o\n", yystate, yychar );
#endif
		if( ++yyps> &yys[YYMAXDEPTH] ) { yyerror( "yacc stack overflow" ); return(1); }
		*yyps = yystate;
		++yypv;
		*yypv = yyval;

 yynewstate:

	yyn = yypact[yystate];

	if( yyn<= YYFLAG ) goto yydefault; /* simple state */

	if( yychar<0 ) if( (yychar=yylex())<0 ) yychar=0;
	if( (yyn += yychar)<0 || yyn >= YYLAST ) goto yydefault;

	if( yychk[ yyn=yyact[ yyn ] ] == yychar ){ /* valid shift */
		yychar = -1;
		yyval = yylval;
		yystate = yyn;
		if( yyerrflag > 0 ) --yyerrflag;
		goto yystack;
		}

 yydefault:
	/* default state action */

	if( (yyn=yydef[yystate]) == -2 ) {
		if( yychar<0 ) if( (yychar=yylex())<0 ) yychar = 0;
		/* look through exception table */

		for( yyxi=yyexca; (*yyxi!= (-1)) || (yyxi[1]!=yystate) ; yyxi += 2 ) ; /* VOID */

		while( *(yyxi+=2) >= 0 ){
			if( *yyxi == yychar ) break;
			}
		if( (yyn = yyxi[1]) < 0 ) return(0);   /* accept */
		}

	if( yyn == 0 ){ /* error */
		/* error ... attempt to resume parsing */

		switch( yyerrflag ){

		case 0:   /* brand new error */

			yyerror( "syntax error" );
		yyerrlab:
			++yynerrs;

		case 1:
		case 2: /* incompletely recovered error ... try again */

			yyerrflag = 3;

			/* find a state where "error" is a legal shift action */

			while ( yyps >= yys ) {
			   yyn = yypact[*yyps] + YYERRCODE;
			   if( yyn>= 0 && yyn < YYLAST && yychk[yyact[yyn]] == YYERRCODE ){
			      yystate = yyact[yyn];  /* simulate a shift of "error" */
			      goto yystack;
			      }
			   yyn = yypact[*yyps];

			   /* the current yyps has no shift onn "error", pop stack */

#ifdef YYDEBUG
			   if( yydebug ) printf( "error recovery pops state %d, uncovers %d\n", *yyps, yyps[-1] );
#endif
			   --yyps;
			   --yypv;
			   }

			/* there is no state on the stack with an error shift ... abort */

	yyabort:
			return(1);


		case 3:  /* no shift yet; clobber input char */

#ifdef YYDEBUG
			if( yydebug ) printf( "error recovery discards char %d\n", yychar );
#endif

			if( yychar == 0 ) goto yyabort; /* don't discard EOF, quit */
			yychar = -1;
			goto yynewstate;   /* try again in the same state */

			}

		}

	/* reduction by production yyn */

#ifdef YYDEBUG
		if( yydebug ) printf("reduce %d\n",yyn);
#endif
		yyps -= yyr2[yyn];
		yypvt = yypv;
		yypv -= yyr2[yyn];
		yyval = yypv[1];
		yym=yyn;
			/* consult goto table to find next state */
		yyn = yyr1[yyn];
		yyj = yypgo[yyn] + *yyps + 1;
		if( yyj>=YYLAST || yychk[ yystate = yyact[yyj] ] != -yyn ) yystate = yyact[yypgo[yyn]];
		switch(yym){
			
case 9:
# line 75 "parse.y"
{
		p_link l;

		l = addlink(yypvt[-2].y_node, yypvt[-1].y_s.ys_node, yypvt[-0].y_cost, yypvt[-1].y_s.ys_net, yypvt[-1].y_s.ys_dir);
		if (GATEWAYED(getnode(yypvt[-1].y_s.ys_node)))
			getlink(l)->l_flag |= LGATEWAY;
		if (yypvt[-1].y_s.ys_flag & TERMINAL)
			getlink(l)->l_flag |= LTERMINAL;
	  } break;
case 10:
# line 84 "parse.y"
{
		p_link l;

		l = addlink(yypvt[-3].y_node, yypvt[-1].y_s.ys_node, yypvt[-0].y_cost, yypvt[-1].y_s.ys_net, yypvt[-1].y_s.ys_dir);
		if (GATEWAYED(getnode(yypvt[-1].y_s.ys_node)))
			getlink(l)->l_flag |= LGATEWAY;
		if (yypvt[-1].y_s.ys_flag & TERMINAL)
			getlink(l)->l_flag |= LTERMINAL;
	  } break;
case 12:
# line 96 "parse.y"
{alias(yypvt[-2].y_node, addnode(yypvt[-0].y_name));} break;
case 13:
# line 97 "parse.y"
{alias(yypvt[-2].y_node, addnode(yypvt[-0].y_name));} break;
case 15:
# line 101 "parse.y"
{fixnet(yypvt[-5].y_node, yypvt[-2].y_node, yypvt[-0].y_cost, DEFNET, DEFDIR);} break;
case 16:
# line 102 "parse.y"
{fixnet(yypvt[-6].y_node, yypvt[-2].y_node, yypvt[-0].y_cost, yypvt[-4].y_net, LRIGHT);} break;
case 17:
# line 103 "parse.y"
{fixnet(yypvt[-6].y_node, yypvt[-3].y_node, yypvt[-0].y_cost, yypvt[-1].y_net, LLEFT);} break;
case 19:
# line 107 "parse.y"
{fixprivate();} break;
case 21:
# line 112 "parse.y"
{yyval.y_node = addnode(yypvt[-0].y_name);} break;
case 22:
# line 113 "parse.y"
{yyval.y_node = addnode("private");} break;
case 23:
# line 114 "parse.y"
{yyval.y_node = addnode("dead");} break;
case 24:
# line 117 "parse.y"
{yyval.y_node = addnode(yypvt[-0].y_name);} break;
case 25:
# line 119 "parse.y"
{
		yyval.y_s.ys_node = addnode(yypvt[-0].y_name);
		yyval.y_s.ys_flag = 0;
	  } break;
case 26:
# line 123 "parse.y"
{
		yyval.y_s.ys_node = addnode(yypvt[-1].y_name);
		yyval.y_s.ys_flag = TERMINAL;
	  } break;
case 27:
# line 129 "parse.y"
{
		yyval.y_s = yypvt[-0].y_s;
		yyval.y_s.ys_net = DEFNET;
		yyval.y_s.ys_dir = DEFDIR;
	  } break;
case 28:
# line 134 "parse.y"
{
		yyval.y_s = yypvt[-0].y_s;
		yyval.y_s.ys_net = yypvt[-1].y_net;
		yyval.y_s.ys_dir = LRIGHT;
	  } break;
case 29:
# line 139 "parse.y"
{
		yyval.y_s = yypvt[-1].y_s;
		yyval.y_s.ys_net = yypvt[-0].y_net;
		yyval.y_s.ys_dir = LLEFT;
	  } break;
case 30:
# line 146 "parse.y"
{yyval.y_node = addprivate(yypvt[-0].y_name);} break;
case 31:
# line 148 "parse.y"
{getnode(yypvt[-0].y_node)->n_flag |= ISPRIVATE;} break;
case 32:
# line 149 "parse.y"
{getnode(yypvt[-0].y_node)->n_flag |= ISPRIVATE;} break;
case 34:
# line 153 "parse.y"
{deadlink(addnode(yypvt[-0].y_name), (p_node) 0);} break;
case 35:
# line 154 "parse.y"
{deadlink(yypvt[-2].y_node, yypvt[-0].y_node);} break;
case 39:
# line 162 "parse.y"
{yyval.y_node = addnode(yypvt[-0].y_name);} break;
case 40:
# line 163 "parse.y"
{
			if (getnode(yypvt[-0].y_node)->n_net == 0) {
				getnode(yypvt[-0].y_node)->n_net = yypvt[-2].y_node;
				yyval.y_node = yypvt[-0].y_node;
			}
	  } break;
case 42:
# line 172 "parse.y"
{yyval.y_cost = DEFCOST;	/* empty -- cost is always optional */} break;
case 43:
# line 173 "parse.y"
{Scanstate = COSTING;} break;
case 44:
# line 173 "parse.y"
{Scanstate = OTHER;} break;
case 45:
# line 174 "parse.y"
{yyval.y_cost = yypvt[-2].y_cost;} break;
case 47:
# line 178 "parse.y"
{yyval.y_cost = yypvt[-1].y_cost;} break;
case 48:
# line 179 "parse.y"
{yyval.y_cost = yypvt[-2].y_cost + yypvt[-0].y_cost;} break;
case 49:
# line 180 "parse.y"
{yyval.y_cost = yypvt[-2].y_cost - yypvt[-0].y_cost;} break;
case 50:
# line 181 "parse.y"
{yyval.y_cost = yypvt[-2].y_cost * yypvt[-0].y_cost;} break;
case 51:
# line 182 "parse.y"
{
		if (yypvt[-0].y_cost == 0)
			yyerror("zero divisor\n");
		else
			yyval.y_cost = yypvt[-2].y_cost / yypvt[-0].y_cost;
	  } break;
		}
		goto yystack;  /* stack new state and value */

	}
